module L2U.Lambda(LmbdExpr(..),lisp2lambda) where

import Text.ParserCombinators.Parsec(parse)
import L2U.LispParser(LispData(..),parser)
import Data.List(delete)

data LmbdExpr = LmAbstr String LmbdExpr
              | LmAppl LmbdExpr LmbdExpr
              | LmVar String deriving (Eq)

instance Show LmbdExpr where
    show (LmAbstr s1 (LmAbstr s2 e)) = show (LmAbstr (s1 ++ " " ++ s2) e)
    show (LmAbstr s e)  = "(\\" ++ s ++ "." ++ show e ++ ")"
    show (LmAppl e1 e2) = "(" ++ show e1 ++ " " ++ show e2 ++ ")"
    show (LmVar s) = s

-- \a b . a
ltrue = LmAbstr "a" $ LmAbstr "b" $ LmVar "a"
-- \a b . b
lfalse = LmAbstr "a" $ LmAbstr "b" $ LmVar "b"
-- a (b ltrue lfalse) lfalse
land a b = LmAppl (LmAppl a (LmAppl (LmAppl b ltrue) lfalse)) lfalse
-- a ltrue (b ltrue lfalse)
lor a b = LmAppl (LmAppl a ltrue) (LmAppl (LmAppl b ltrue) lfalse)
-- a lfalse ltrue
lnot a = LmAppl (LmAppl a lfalse) ltrue

-- \pb . pb a b
lpair a b = LmAbstr "[pb]" $ LmAppl (LmAppl (LmVar "[pb]") a) b
-- x ltrue
lfst x = LmAppl x ltrue
-- x lfalse
lsnd x = LmAppl x lfalse

-- \a . a
lid = LmAbstr "a" $ LmVar "a"
-- \a . x
lconst x = LmAbstr "[a]" x

-- TODO: uwaga, mozliwosc przechwycenia zmiennych!
-- n -> \f x . f (f ... (f x))
lnat n = LmAbstr "f" $ LmAbstr "x" $ hlp n where
    hlp 0 = LmVar "x"
    hlp n = LmAppl (LmVar "f") (hlp $ n-1)
lnat_inc n = LmAbstr "f" $ LmAbstr "x" $ LmAppl (LmVar "f") 
    (LmAppl (LmAppl n (LmVar "f")) (LmVar "x"))
lnat_add n m = LmAbstr "f" $ LmAbstr "x" $ 
    LmAppl (LmAppl n (LmVar "f")) (LmAppl (LmAppl m (LmVar "f")) (LmVar "x"))
lnat_dec n = lsnd $ LmAppl (LmAppl n (LmAbstr "y" $ 
        lpair (lnat_inc $ lfst $ LmVar "y") (lfst $ LmVar "y"))) 
    (lpair (lnat 0) (lnat 0))
lnat_sub n m = LmAppl (LmAppl m (LmAbstr "y" $ lnat_dec $ LmVar "y")) n
lnat_iszero n = LmAppl (LmAppl n (LmAbstr "x" lfalse)) ltrue
lnat_iseven n = LmAppl (LmAppl n (LmAbstr "x" $ lnot (LmVar "x"))) ltrue
lint n | n < 0     = lpair (lnat 0) (lnat (-n))
       | otherwise = lpair (lnat n) (lnat 0)
lint_inc n = LmAppl (LmAbstr "n" $ lpair (lnat_inc $ lfst $ LmVar "n") 
                                         (lsnd $ LmVar "n")) n
lint_dec n = LmAppl (LmAbstr "n" $ lpair (lfst $ LmVar "n") 
                                         (lnat_inc $ lsnd $ LmVar "n")) n
lint_add n m = LmAppl (LmAppl (LmAbstr "n" $ LmAbstr "m" $ 
    lpair (lnat_add (lfst $ LmVar "n") (lfst $ LmVar "m")) 
          (lnat_add (lsnd $ LmVar "n") (lsnd $ LmVar "m"))) n) m
lint_sub n m = LmAppl (LmAppl (LmAbstr "n" $ LmAbstr "m" $ 
    lpair (lnat_add (lfst $ LmVar "n") (lsnd $ LmVar "m")) 
          (lnat_add (lsnd $ LmVar "n") (lfst $ LmVar "m"))) n) m
lint_iszero n = LmAppl (LmAbstr "n" $
    land (lnat_iszero $ lnat_sub (lfst $ LmVar "n") (lsnd $ LmVar "n"))
         (lnat_iszero $ lnat_sub (lsnd $ LmVar "n") (lfst $ LmVar "n"))) n
lint_ispos n = LmAppl (LmAbstr "n" $
    lnot $ lnat_iszero $ lnat_sub (lfst $ LmVar "n") (lsnd $ LmVar "n")) n
lint_isneg n = LmAppl (LmAbstr "n" $
    lnot $ lnat_iszero $ lnat_sub (lsnd $ LmVar "n") (lfst $ LmVar "n")) n
lint_iseven n = LmAppl (LmAbstr "n" $ 
    lnat_iseven $ lnat_add (lfst $ LmVar "n") (lsnd $ LmVar "n")) n
lint_eq n m = lint_iszero $ lint_sub n m
lint_gt n m = lint_ispos $ lint_sub n m

[tcons,tnil,tatom,tnum,tstr,tbool,tfun] = map tany [0..6] where
    tany i = foldr LmAbstr (LmVar $ l !! i) l 
    l = ["a","b","c","d","e","f","g"]

[if2tcons,if2tnil,if2tatom,if2tnum,if2tstr,if2tbool,if2tfun] = 
    map tif [0..6] where
    tif i e f g = LmAppl (LmAppl (LmAppl good (LmAbstr "[hh]" $ LmAppl f (lsnd e))) (LmAbstr "[hh]" $ LmAppl g (lsnd e))) lid where
        good = foldl LmAppl (lfst e) 
                       ((take i falses) ++ [ltrue] ++ (take (6-i) falses))
    falses = repeat lfalse

[iftcons,iftnil,iftatom,iftnum,iftstr,iftbool,iftfun] = 
    map tif [if2tcons,if2tnil,if2tatom,if2tnum,if2tstr,if2tbool,if2tfun]
    where tif x e f = x e f verror

typedv t v = lpair t v

vtrue  = typedv tbool ltrue
vfalse = typedv tbool lfalse
isvtrue x = if2tbool x lid (lconst ltrue)
vand a b = typedv tbool $ land (isvtrue a) (isvtrue b) 
vor a b = typedv tbool $ lor (isvtrue a) (isvtrue b)
vnot a = typedv tbool $ lnot (isvtrue a)
vcons a b = typedv tcons (lpair a b)
vnil = typedv tnil lid
vfun f = typedv tfun f
vnum n = typedv tnum (lint n)
vadd n m = typedv tnum $ lint_add (iftnum n lid) (iftnum m lid)
vsub n m = typedv tnum $ lint_sub (iftnum n lid) (iftnum m lid)
viszero n = typedv tbool $ lint_iszero $ iftnum n lid
vispos n = typedv tbool $ lint_ispos $ iftnum n lid
visneg n = typedv tbool $ lint_isneg $ iftnum n lid
viseven n = typedv tbool $ lint_iseven $ iftnum n lid
visodd n = typedv tbool $ lnot $ lint_iseven $ iftnum n lid
vnumeq n m = typedv tbool $ lint_eq (iftnum n lid) (iftnum m lid)
vnumgt n m = typedv tbool $ lint_gt (iftnum n lid) (iftnum m lid)
verror = LmAppl (LmVar "\0E") (LmAppl (LmVar "\0.q") (LmVar "\0I"))

--vcar e = iftcons e (LmAbstr "x" $ lfst (LmVar "x"))
--vcdr e = iftcons e (LmAbstr "x" $ lsnd (LmVar "x"))
vcar e = lfst $ iftcons e lid
vcdr e = lsnd $ iftcons e lid

stdlib = [
    ("pair?", lambdah ["p"] False (checker if2tcons)),
    ("null?", lambdah ["p"] False (checker if2tnil)),
    ("boolean?", lambdah ["p"] False (checker if2tbool)),
    ("symbol?", lambdah ["p"] False (checker if2tatom)),
    ("number?", lambdah ["p"] False (checker if2tnum)),
    ("integer?", lambdah ["p"] False (checker if2tnum)),
    ("procedure?", lambdah ["p"] False (checker if2tfun)),
    ("cons", lambdah ["h","t"] False (vcons (LmVar "h") (LmVar "t"))),
    ("car", lambdah ["p"] False (vcar (LmVar "p"))),
    ("cdr", lambdah ["p"] False (vcdr (LmVar "p"))),
    ("and_", lambdah ["a","b"] False (vand (LmVar "a") (LmVar "b"))),
    ("or_", lambdah ["a","b"] False (vor (LmVar "a") (LmVar "b"))),
    ("not", lambdah ["a"] False (vnot (LmVar "a"))),
    ("add_", lambdah ["a","b"] False (vadd (LmVar "a") (LmVar "b"))),
    ("sub_", lambdah ["a","b"] False (vsub (LmVar "a") (LmVar "b"))),
    ("zero?", lambdah ["a"] False (viszero (LmVar "a"))),
    ("positive?", lambdah ["a"] False (vispos (LmVar "a"))),
    ("negative?", lambdah ["a"] False (visneg (LmVar "a"))), 
    ("odd?", lambdah ["a"] False (visodd (LmVar "a"))),
    ("even?", lambdah ["a"] False (viseven (LmVar "a"))),
    ("eq_", lambdah ["a","b"] False (vnumeq (LmVar "a") (LmVar "b"))),
    ("gt_", lambdah ["a","b"] False (vnumgt (LmVar "a") (LmVar "b"))) 
    ] ++ derived where
    checker x = (x (LmVar "p") (LmAbstr "" vtrue) (LmAbstr "" vfalse))
    derived = map pdervd stdlib_derived
    pdervd (n,t) = case parse parser "" t of Right e -> (n,l2l_code e)

stdlib_derived = [
    ("cadr", "(lambda (l) (car (cdr l)))"),
    ("list", "(lambda l l)"),
    ("list?", "(letrec ((list? (lambda (l) (cond ((pair? l) (list? (cdr l))) ((null? l) #t) (else #f))))) list?)"),
    ("foldl", "(letrec ((foldl (lambda (f a l) (cond ((pair? l) (foldl f (f a (car l)) (cdr l))) ((null? l) a) (else (error)))))) foldl)"),
    ("foldr", "(letrec ((foldr (lambda (f a l) (cond ((pair? l) (f (car l) (foldr f a (cdr l)))) ((null? l) a) (else (error)))))) foldr)"),
    ("append_", "(letrec ((append2 (lambda (a b) (cond ((null? a) b) ((pair? a) (cons (car a) (append2 (cdr a) b))) (else (error)))))) append2)"),
    ("append", "(lambda l (foldr append_ '() l))"),
    ("reverse", "(lambda (l) (foldl (lambda (a b) (cons b a)) '() l))"),
    ("and", "(lambda l (foldl and_ #t l))"),
    ("or", "(lambda l (foldl or_ #f l))"),
    ("+", "(lambda l (foldl add_ 0 l))"),
    ("-", "(lambda l (foldl sub_ (car l) (cdr l)))"), -- TODO minus unarny
    ("multirel_", "(letrec ((multirel (lambda (f l) (cond ((null? l) #t) ((null? (cdr l)) #t) (else (and (f (car l) (cadr l)) (multirel f (cdr l)))))))) multirel)"),
    ("=", "(lambda l (multirel_ eq_ l))"),
    (">", "(lambda l (multirel_ gt_ l))"),
    ("<", "(lambda l (multirel_ (lambda (a b) (gt_ b a)) l))")
    ]
--    ("append", "(letrec ((appendl (lambda (l) (cond ((null? l) '()) ((pair? l) (append_ (car l) (appendl (cdr l)))))))) (lambda l (appendl l)))")]
--    ("length", "(letrec ((length (lambda (l) (cond ((pair? l) (+ 1 (length (cdr l)))) ((null? l) 0) (else (error)))))) length)"),

l2l_quoted (LBool True)  = vtrue
l2l_quoted (LBool False) = vfalse
l2l_quoted (LCons a b)   = vcons (l2l_quoted a) (l2l_quoted b)
l2l_quoted LNil          = vnil
l2l_quoted (LNum n)      = vnum n

l2l_qquoted (LCons (LAtom "unquote") (LCons x LNil)) = l2l_code x
l2l_qquoted (LCons a b) = vcons (l2l_qquoted a) (l2l_qquoted b)
l2l_qquoted x = l2l_quoted x

lambdah flist tll body = vfun $ outside flist tll where
    inside = foldr LmAbstr body (reverse flist)
    outside [] False = LmAbstr "" $ iftnil (LmVar "") (LmAbstr "" $ inside)
    outside [_] True = LmAbstr "" $ LmAppl inside (LmVar "")
    outside (_:t) x  = LmAbstr "" $ LmAppl (LmAppl (outside t x) 
                                           (vcdr $ LmVar "")) (vcar $ LmVar "")

l2l_lambda formals body = lambdah flist tll (l2l_begin body) where
    scan_formals LNil      = ([],False)
    scan_formals (LAtom t) = ([t],True)
    scan_formals (LCons (LAtom s) r) = (s:l,t) where (l,t) = scan_formals r
    (flist,tll) = scan_formals formals

letbinds :: LispData -> [(String,LispData)]
letbinds LNil = []
letbinds (LCons (LCons (LAtom a) (LCons c LNil)) r) = (a,c):letbinds r

l2l_let b body = foldl ahlp (foldr lhlp (l2l_begin body) lb) lb where
    lhlp (a,_) ac = LmAbstr a ac
    ahlp ac (_,c) = LmAppl ac (l2l_code c)
    lb = letbinds b

l2l_lets b body = foldr hlp (l2l_begin body) (letbinds b) where
    hlp (a,c) ac = LmAppl (LmAbstr a ac) (l2l_code c)

l2l_letrec b body = foldl ahlp (fpack $ lsubst lba $ l2l_begin body) lb where
    lhlp a ac = LmAbstr a ac
    ahlp ac (_,c) = LmAppl ac (fpack $ lsubst lba $ l2l_code c)
    shlp ac v = LmAppl ac (LmVar v)
    fpack c = foldr lhlp c lba
    lb = letbinds b
    lba = map fst lb
    lsubst [] e = e
    lsubst m (LmAbstr x e) = LmAbstr x (lsubst (delete x m) e)
    lsubst m (LmAppl e1 e2) = LmAppl (lsubst m e1) (lsubst m e2)
    lsubst m (LmVar x) | x `elem` m = foldl shlp (LmVar x) lba
                       | otherwise  = LmVar x

l2l_begin LNil = verror
l2l_begin (LCons e LNil) = l2l_code e
l2l_begin (LCons e r) = LmAppl (LmAbstr "" $ l2l_begin r) (l2l_code e)

-- complicated to make it lazy
l2l_if t c a = LmAppl dowork (isvtrue $ l2l_code t) where
    dowork = (LmAbstr "" $ LmAppl ((LmAppl (LmAppl (LmVar "") (LmAbstr "" $ l2l_code c)) (LmAbstr "" $ l2l_code a))) lid)

l2l_cond LNil = vfalse
l2l_cond (LCons (LCons (LAtom "else") c) LNil) = l2l_begin c
l2l_cond (LCons (LCons t c) r) = 
    l2l_if t (LCons (LAtom "begin") c) (LCons (LAtom "cond") r)

l2l_appl op args = iftfun (l2l_code op) dowork where
    dowork = (LmAbstr "[f]" $ LmAppl (LmVar "[f]") (fargs args))
    fargs LNil        = vnil
    fargs (LCons a r) = (vcons (l2l_code a) (fargs r))

l2l_error = verror

l2l_print [] ac = LmAppl (LmVar "\0R") ac
l2l_print (c:r) ac = l2l_print r (LmAppl (LmVar ['\0','.',c]) ac)

l2l_code LNil            = verror
l2l_code (LAtom s)       = LmVar s
l2l_code (x @ (LBool _)) = l2l_quoted x
l2l_code (x @ (LNum _))  = l2l_quoted x
l2l_code (x @ (LStr _))  = l2l_quoted x
l2l_code (LCons (LAtom "error") _) = verror -- nie! error ma byc procedura
l2l_code (LCons (LAtom "print") (LCons (LStr s) LNil)) = 
    l2l_print s (LmVar "\0I")
l2l_code (LCons (LAtom "begin") body) = l2l_begin body
l2l_code (LCons (LAtom "quote") (LCons e LNil)) = l2l_quoted e
l2l_code (LCons (LAtom "quasiquote") (LCons e LNil)) = l2l_qquoted e
l2l_code (LCons (LAtom "lambda") (LCons form body)) = 
    l2l_lambda form body
l2l_code (LCons (LAtom "if") (LCons test (LCons cons LNil))) = 
    l2l_if test cons (LBool False)
l2l_code (LCons (LAtom "if") (LCons test (LCons cons (LCons alt LNil)))) = 
    l2l_if test cons alt
l2l_code (LCons (LAtom "cond") clauses) = l2l_cond clauses
l2l_code (LCons (LAtom "let") (LCons bnd body)) =
    l2l_let bnd body
l2l_code (LCons (LAtom "let*") (LCons bnd body)) =
    l2l_lets bnd body
l2l_code (LCons (LAtom "letrec") (LCons bnd body)) =
    l2l_letrec bnd body
l2l_code (LCons op args) = l2l_appl op args

lisp2lambda :: LispData -> LmbdExpr
lisp2lambda e = foldr (\(n,c) a -> LmAppl (LmAbstr n a) c) (l2l_code e) stdlib
--lisp2lambda e = l2l_code e

