module L2U.LispParser(LispData(..),parser) where

import Text.ParserCombinators.Parsec
import Control.Monad
import Data.Maybe

data LispData = LCons LispData LispData
              | LNil
              | LAtom String 
              | LNum Integer
              | LStr String
              | LBool Bool deriving (Show,Eq)

symbol :: Parser Char
symbol = oneOf "!$#%&|*+-/:<=>?@^_~"

parseString :: Parser LispData
parseString = do
    char '"'
    x <- many (noneOf "\"")
    char '"'
    return $ LStr x

parseAtom :: Parser LispData
parseAtom = do
    first <- letter <|> symbol
    rest  <- many (letter <|> digit <|> symbol)
    return $ case first:rest of
        "#t" -> LBool True
        "#f" -> LBool False
        x    -> LAtom x

parseNumber :: Parser LispData
parseNumber = do 
    mn  <- option ' ' (char '-')
    dgt <- many1 digit
    return $ LNum $ read (mn:dgt)

parseList :: Parser LispData
parseList = do
    char '('
    spaces
    head <- endBy parseExpr spaces
    tail <- option LNil (char '.' >> spaces >> parseExpr)
    spaces
    char ')'
    return $ foldr LCons tail head

parseQuoted :: Parser LispData
parseQuoted = do
    c <- choice (map string ["'","`",",",",@"])
    x <- parseExpr
    return $ LCons (LAtom $ fromJust $ lookup c l) $ LCons x LNil
    where l = [("'","quote"),("`","quasiquote"),(",","unquote"),
               (",@","unquote-splicing")]

parseExpr = try parseNumber
        <|> parseAtom
        <|> parseString
        <|> parseQuoted
        <|> parseList

parser = parseExpr

