module L2U.Unlambda(UnCode(..),lambda2unlambda) where

import L2U.Lambda

data UnCode = UnAppl UnCode UnCode | UnPrint Char | UnR
            | UnK | UnS | UnI | UnV | UnC | UnE deriving Eq

instance Show UnCode where
    show UnK = "k"
    show UnS = "s"
    show UnI = "i"
    show UnR = "r"
    show UnV = "v"
    show UnE = "e"
    show UnC = "c"
    show (UnAppl e1 e2) = "`" ++ show e1 ++ show e2
    show (UnPrint c)    = ['.',c]

hasvar :: LmbdExpr -> String -> Bool
hasvar (LmVar x) y     | x == y = True
hasvar (LmAbstr x e) y | x == y = False
hasvar (LmAbstr x e) y          = hasvar e y
hasvar (LmAppl e1 e2) y         = hasvar e1 y || hasvar e2 y
hasvar _ _                      = False

sideeft :: LmbdExpr -> Bool
sideeft (LmVar _) = False
sideeft (LmAbstr _ e) = sideeft e
sideeft (LmAppl (LmVar "\0K") e) = sideeft e
sideeft (LmAppl (LmVar "\0D") e) = sideeft e
sideeft (LmAppl (LmVar "\0S") e) = sideeft e
sideeft (LmAppl (LmAppl (LmVar "\0S") e1) e2) = sideeft e1 || sideeft e2
sideeft (LmAppl _ _) = True

l2u (LmVar x) = LmVar x
l2u (LmAppl e1 e2) = LmAppl (l2u e1) (l2u e2)
l2u (LmAbstr x e) | not (hasvar e x || sideeft tr) = 
    LmAppl (LmVar "\0K") tr where tr = l2u e
l2u (LmAbstr x (LmVar y)) | x /= y = LmAppl (LmVar "\0K") (LmVar y)
l2u (LmAbstr x (LmVar y)) | x == y = LmVar "\0I"
l2u (LmAbstr x (LmAppl e1 e2)) = LmAppl (LmAppl (LmVar "\0S") 
                                                (l2u $ LmAbstr x e1)) 
                                        (l2u $ LmAbstr x e2)
l2u (LmAbstr x (LmAbstr y e)) = l2u (LmAbstr x $ l2u $ LmAbstr y e)

lh (LmVar "\0K")  = UnK
lh (LmVar "\0I")  = UnI
lh (LmVar "\0S")  = UnS
lh (LmVar "\0R")  = UnR
lh (LmVar "\0V")  = UnV
lh (LmVar "\0C")  = UnC
lh (LmVar "\0E")  = UnE
lh (LmVar ('\0':'.':[c]))  = UnPrint c
lh (LmVar v)      = error ("free variable: " ++ v)
lh (LmAppl e1 e2) = UnAppl (lh e1) (lh e2)
lh (LmAbstr _ _)  = error "abstraction"

lambda2unlambda = lh . l2u

